/* parse_cmdline.c - parsing of command line options */

#include "common_func.h" /* should be included before the C library files */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <ctype.h>  /* isspace */
#include <sys/stat.h> /* stat() */
#ifdef _WIN32
#include <windows.h> /* for SetFileApisToOEM(), CharToOem() */
#endif

#include "librhash/rhash.h"
#include "librhash/plug_openssl.h"
#include "win_utils.h"
#include "file_mask.h"
#include "crc_print.h"
#include "output.h"
#include "rhash_main.h"
#include "version.h"
#include "parse_cmdline.h"

#define VERSION_STRING PROGRAM_NAME " v" VERSION "\n"

typedef struct options_t options_t;
struct options_t conf_opt; /* config file parsed options */
struct options_t opt;      /* command line options */

/**
 * Print program help.
 */
static void print_help(void)
{
  puts( VERSION_STRING "\n"
    "Usage: " CMD_FILENAME " [<option> ...] <filename|-> [...]\n"
    "       " CMD_FILENAME " --printf=<format string> <filename|-> [...]\n\n"
    "Options:\n"
    "  -V, --version Print " PROGRAM_NAME " version and exit.\n"
    "  -h, --help    Print this help screen.\n"
    "  -C, --crc32   Calculate CRC32 hash sum.\n"
    "  -M, --md5     Calculate MD5   hash sum.\n"
    "  -H, --sha1    Calculate SHA1  hash sum.\n"
    "      --sha224, --sha256, --sha384, --sha512 Calculate SHA2 hash sum.\n"
    "  -G, --tiger   Calculate Tiger hash sum.\n"
    "  -T, --tth     Calculate TTH sum.\n"
    "      --btih    Calculate BitTorrent InfoHash.\n"
    "  -A, --aich    Calculate AICH hash.\n"
    "  -E, --ed2k    Calculate eDonkey hash sum.\n"
    "  -L, --ed2k-link  Calculate and print eDonkey link.\n"
    "      --gost    Calculate GOST R 34.11-94 hash.\n"
    "      --gost-cryptopro CryptoPro version of the GOST R 34.11-94 hash.\n"
    "  -a, --all     Print all supported hash sums.\n"
    "  -c, --check   Check crc files specified by command line.\n"
    "  -u, --update  Update crc files specified by command line.\n"
    "  -e, --embed-crc  Rename files by inserting crc32 sum into name.\n"
    "      --check-embedded  Verify files by crc32 sum embedded in their names.\n"
    "  -B, --benchmark  Benchmark selected algorithm.\n"
    "  -v, --verbose Be verbose.\n"
    "  -r, --recursive  Process directories recursively.\n"
    "      --skip-ok Don't print OK messages for successfully verified files.\n"
    "  -i, --ignore-case  Ignore case of filenames when updating crc files.\n"
    "      --percents   Show percents, while calculating or checking sums.\n"
    "      --speed   Print per-file and total processing speed.\n"
    "  -q, --accept=<list> Comma-delimited list of extensions of files\n"
    "                to calculate sums for. Used with -r.\n"
    "  -t, --crc-accept=<list> Comma-delimited list of crc file extensions.\n"
    "      --maxdepth=<n> Descend at most <n> levels of directories below\n"
    "                the command line arguments. The `--maxdepth 0' means\n"
    "                hash and check only the command line arguments.\n"
    "  -o, --output=<file> File to output calculation or checking results.\n"
    "  -l, --log=<file>    File to log errors and verbose information.\n"
    "      --sfv     Print hash sums, using SFV format (default).\n"
    "      --bsd     Print hash sums, using BSD-like format.\n"
    "      --simple  Print hash sums, using simple format.\n"
    "  -m, --magnet  Print hash sums  as magnet links.\n"
    "      --torrent Create torrent files.\n"
#ifdef _WIN32
    "      --ansi    Use Windows codepage for output (Windows only).\n"
#endif
    "      --template=<file> Load a printf-like template from the <file>\n"
    "  -p, --printf=<format string>  Format and print hash sums.\n"
    "                See man page for details.\n"
  );
  rsh_exit(0);
}

/**
 * Print the names of supported hash algorithms to console.
 */
static void list_hashes(void)
{
  int id;
  for(id = 1; id < RHASH_ALL_HASHES; id <<= 1) {
    const char* hash_name = rhash_get_name(id);
    if(hash_name) printf("%s\n", hash_name);
  }
  rsh_exit(0);
}

/**
 * Process --accept and --crc-accept options.
 */
static void crc_accept(options_t *o, char* accept_string, unsigned type)
{
  file_mask_array** ptr = (type ? &o->crc_accept : &o->files_accept);
  if(!*ptr) *ptr = file_mask_new();
  file_mask_add_list(*ptr, accept_string);
}

/**
 * Process an --openssl option.
 */
static void openssl_flags(options_t *o, char* openssl_hashes, unsigned type)
{
#ifdef USE_OPENSSL
  char *cur, *next;
  (void)type;
  o->openssl_mask = 0x80000000;

  /* set the openssl_mask */
  for(cur = openssl_hashes; cur && *cur; cur = next) {
    print_hash_info *info = hash_info_table;
    unsigned bit;
    size_t length;
    next = strchr(cur, ',');
    length = (next != NULL ? (size_t)(next - cur) : strlen(cur));

    for(bit = 1; bit <= RHASH_ALL_HASHES; bit = bit << 1, info++) {
      if( (bit && OPENSSL_SUPPORTED_HASHES_MASK) &&
          memcmp(cur, info->short_name, length) == 0 &&
          info->short_name[length] == 0) {
        o->openssl_mask |= bit;
        break;
      }
    }
    if(bit > RHASH_ALL_HASHES) {
      cur[length] = 0;
      log_msg(PROGRAM_NAME " warning: openssl option doesn't support '%s' hash\n", cur);
    }
  }
#else
  (void)type;
  (void)openssl_hashes;
  (void)o;
  log_msg(PROGRAM_NAME " warning: compiled without openssl support\n");
#endif
}

/**
 * Process --video option.
 */
static void accept_video(options_t *o)
{
  crc_accept(o, ".avi,.ogm,.mkv,.mp4,.mpeg,.mpg,.asf,.rm,.wmv,.vob", 0);
}

/**
 * Say nya! Keep secret! =)
 */
static void nya(void)
{
  printf("\n  /\\___/\\   _\n = ' T ' = //\n"
    "  |     |\\//\n  | | | | )\n  \"\"  \"\" \"\n");
  rsh_exit(0);
}

/**
 * Process on --maxdepth option.
 *
 * @param o pointer to the processed option
 * @param number string containing the max-depth number
 * @param param unused parameter
 */
static void set_max_depth(options_t *o, char* number, unsigned param)
{
  (void)param;
  if(strspn(number, "0123456789") < strlen(number)) {
    log_msg(PROGRAM_NAME ": maxdepth parameter is not a number: %s\n", number);
    rsh_exit(2);
  }
  o->find_max_depth = atoi(number);
}

/**
 * Set bittorent file piece-length
 *
 * @param o pointer to the processed option
 * @param number string containing the bt-piece-length number
 * @param param unused parameter
 */
static void set_bt_piece_length(options_t *o, char* number, unsigned param)
{
  (void)param;
  if(strspn(number, "0123456789") < strlen(number)) {
    log_msg(PROGRAM_NAME ": bt-piece-length parameter is not a number: %s\n", number);
    rsh_exit(2);
  }
  o->bt_piece_length = (size_t)atoi(number);
}

/**
 * Set the path separator to use when printing paths
 *
 * @param o pointer to the processed option
 * @param sep file separator, can be only '/' or '\'
 * @param param unused parameter
 */
static void set_path_separator(options_t *o, char* sep, unsigned param)
{
  (void)param;
  if((*sep == '/' || *sep == '\\') && sep[1] == 0) {
    o->path_separator = *sep;
#if defined(_WIN32)
  /* MSYS enviropment changes '/' in command line to HOME, see http://www.mingw.org/wiki/FAQ */
  } else if(getenv("MSYSTEM") || getenv("TERM")) {
    log_msg("warning: wrong path-separator, use '//' instead of '/' on MSYS\n");
    o->path_separator = '/';
#endif
  } else {
    log_msg(PROGRAM_NAME ": error: path-separator is not '/' or '\\': %s\n", sep);
    rsh_exit(2);
  }
}

/**
 * Information about a console option of the program.
 */
typedef struct cmdline_opt_t
{
  char  type;
  char  short1, short2;
  char* long_name;
  void* ptr;
  unsigned param;
} cmdline_opt_t;

/* supported program options */
cmdline_opt_t cmdline_opt[] = {
  /* program modes */
  { 'u', 'c',   0, "check",  &opt.mode, MODE_CHECK },
  { 'u',   0,   0, "check-embedded",  &opt.mode, MODE_CHECK_EMBEDDED },
  { 'u', 'u',   0, "update", &opt.mode, MODE_UPDATE },
  { 'u', 'B',   0, "benchmark", &opt.mode, MODE_BENCHMARK },
  { 'u',   0,   0, "torrent", &opt.mode, MODE_TORRENT },
  /* hash sums options */
  { 'u', 'a',   0, "all",    &opt.sum_flags, RHASH_ALL_HASHES },
  { 'u', 'C',   0, "crc32",  &opt.sum_flags, RHASH_CRC32 },
  { 'u',   0,   0, "md4",    &opt.sum_flags, RHASH_MD4 },
  { 'u', 'M',   0, "md5",    &opt.sum_flags, RHASH_MD5 },
  { 'u', 'H',   0, "sha1",   &opt.sum_flags, RHASH_SHA1 },
  { 'u',   0,   0, "sha224", &opt.sum_flags, RHASH_SHA224 },
  { 'u',   0,   0, "sha256", &opt.sum_flags, RHASH_SHA256 },
  { 'u',   0,   0, "sha384", &opt.sum_flags, RHASH_SHA384 },
  { 'u',   0,   0, "sha512", &opt.sum_flags, RHASH_SHA512 },
  { 'u', 'G',   0, "tiger",  &opt.sum_flags, RHASH_TIGER },
  { 'u', 'T',   0, "tth",    &opt.sum_flags, RHASH_TTH },
  { 'u',   0,   0, "btih",   &opt.sum_flags, RHASH_BTIH },
  { 'u', 'E',   0, "ed2k",   &opt.sum_flags, RHASH_ED2K },
  { 'u', 'A',   0, "aich",   &opt.sum_flags, RHASH_AICH },
  { 'u',   0,   0, "gost",   &opt.sum_flags, RHASH_GOST },
  { 'u',   0,   0, "gost-cryptopro", &opt.sum_flags, RHASH_GOST_CRYPTOPRO },
  { 'u', 'W',   0, "whirlpool", &opt.sum_flags, RHASH_WHIRLPOOL },
  { 'u',   0,   0, "ripemd160", &opt.sum_flags, RHASH_RIPEMD160 },
  { 'u',   0,   0, "has160",    &opt.sum_flags, RHASH_HAS160 },
  { 'u',   0,   0, "snefru128", &opt.sum_flags, RHASH_SNEFRU128 },
  { 'u',   0,   0, "snefru256", &opt.sum_flags, RHASH_SNEFRU256 },
  { 'u',   0,   0, "edonr256",  &opt.sum_flags, RHASH_EDONR256 },
  { 'u',   0,   0, "edonr512",  &opt.sum_flags, RHASH_EDONR512 },
  { 'u', 'L',   0, "ed2k-link", &opt.sum_flags, OPT_ED2K_LINK },
  /* output formats */
  { 'u',   0,   0, "sfv",     &opt.fmt, FMT_SFV },
  { 'u',   0,   0, "bsd",     &opt.fmt, FMT_BSD },
  { 'u',   0,   0, "simple",  &opt.fmt, FMT_SIMPLE },
  { 'u', 'm',   0, "magnet",  &opt.fmt, FMT_MAGNET },
  { 'u',   0,   0, "uppercase", &opt.flags, OPT_UPPERCASE },
  { 'u',   0,   0, "lowercase", &opt.flags, OPT_LOWERCASE },
  { 's', 'p',   0, "printf", &opt.printf, 0 },
  { 's',   0,   0, "template", &opt.template_file, 0 },
  /* other options */
  { 'u', 'r', 'R', "recursive", &opt.flags, OPT_RECURSIVE },
  { 'u', 'v',   0, "verbose", &opt.flags, OPT_VERBOSE },
  { 'u',   0,   0, "gost-reverse", &opt.flags, OPT_GOST_REVERSE },
  { 'u',   0,   0, "skip-ok", &opt.flags, OPT_SKIP_OK },
  { 'u', 'i',   0, "ignore-case", &opt.flags, OPT_IGNORE_CASE },
  { 'u',   0,   0, "percents", &opt.flags, OPT_PERCENTS },
  { 'u',   0,   0, "speed",  &opt.flags, OPT_SPEED },
  { 'u', 'e',   0, "embed-crc",  &opt.flags, OPT_EMBED_CRC },
  { 's',   0,   0, "embed-crc-delimiter", &opt.embed_crc_delimiter, 0 },
  { 'F',   0,   0, "path-separator", set_path_separator, 0 },
  { 's', 'o',   0, "output", &opt.output, 0 },
  { 's', 'l',   0, "log",    &opt.log,    0 },
  { 'F', 'q',   0, "accept", crc_accept, 0 },
  { 'F', 't',   0, "crc-accept", crc_accept, 1 },
  { 'f',   0,   0, "video",  accept_video, 0 },
  { 'f',   0,   0, "nya",  nya, 0 },
  { 'F',   0,   0, "maxdepth", set_max_depth, 0 },
  { 'u',   0,   0, "bt-private", &opt.flags, OPT_BT_PRIVATE },
  { 'F',   0,   0, "bt-piece-length", set_bt_piece_length, 0 },
  { 's',   0,   0, "bt-announce", &opt.bt_announce, 0 },
  { 'p', 'V',   0, "version", VERSION_STRING, 0 },
  { 'f', 'h',   0, "help",   print_help, 0 },
  { 'f',   0,   0, "list-hashes", list_hashes, 0 },
  { 'u',   0,   0, "benchmark-raw", &opt.flags, OPT_BENCH_RAW },
  { 'F',   0,   0, "openssl", openssl_flags, 0 },
#ifdef _WIN32 /* code pages */
  { 'u',   0,   0, "utf8", &opt.flags, OPT_UTF8 },
  { 'u',   0,   0, "ansi", &opt.flags, OPT_ANSI },
  { 'u',   0,   0, "oem",  &opt.flags, OPT_OEM },
#endif
  { 0,0,0,0,0,0 }
};

/**
 * Return non-zero if given option has a required parameter
 *
 * @param type the type of option to check
 */
static int is_param_required(char type)
{
  return (type == 's' || type == 'F');
}

/* Windows-only functions to detect encoding before processing other cmd-line arguments */
#ifdef _WIN32

/**
 * Test if given option specifies program encoding.
 *
 * @return the encodig flag on success, 0 otherwise
 */
static int test_encoding_option(const wchar_t* wname)
{
  static const wchar_t* warr[3] = { L"utf8", L"ansi", L"oem" };
  static const int opt_arr[3] = { OPT_UTF8, OPT_ANSI, OPT_OEM };
  int i;
  for(i = 0; i < 3 && wcscmp(wname, warr[i]) != 0; i++);
  return (i < 3 ? opt_arr[i] : 0);
}

/**
 * Detect encoding from command line options.
 *
 * @param wargv
 * @param nArg
 * @return the encodig flag on success, 0 otherwise
 */
int detect_encoding(wchar_t** wargv, int nArg)
{
  int i, d = 0, res = 0;
  for(i = 1, wargv++; i < nArg && !res; i++, wargv++) {
    /* if not an option, assume it is an option argument */
    if((*wargv)[0] != L'-' || (*wargv)[1] == 0) {
      if( (++d) >= 2) break;
      continue;
    }
    d = 0;
    if((*wargv)[1] != L'-') continue;
    if((*wargv)[2] == 0) break;
    res = test_encoding_option((*wargv) + 2);
  }
  if(!res) res = (conf_opt.flags & OPT_ENCODING);
  return (res ? res : OPT_UTF8);
}
#endif

/**
 * Process given command line option
 *
 * @param o option to process
 * @param value the value for the option, if required
 */
static void process_option_2(options_t *opts, cmdline_opt_t *o, char* value)
{
  if(o->type == 'u') {
    *(unsigned*)((char*)opts + ((char*)o->ptr - (char*)&opt)) |= o->param;
  } else if(o->type == 's' || o->type == 'F') {
    /* value is required */
    if(!value) {
      log_msg(PROGRAM_NAME ": argument required for option %s\n", o->long_name);
      rsh_exit(2);
    }
    if(o->type == 's') {
      /* save option argument */
      *(char**)((char*)opts + ((char*)o->ptr - (char*)&opt)) = value;
    } else {
      /* call option argument handler */
      ( ( void(*)(options_t *, char*, unsigned) )o->ptr )(opts, value, o->param);
    }
  } else if(o->type == 'f') {
    /* call option handler */
    ( ( void(*)(options_t *) )o->ptr )(opts);
  } else if(o->type == 'p') {
    puts((char*)o->ptr);
    rsh_exit(0);
  }
}

/**
 * Process given command line option
 *
 * @param is_short non zero if option is in its short form
 * @param parg pointer to the first unprocessed program argument
 * @param popt option name
 */
static char* retrive_option_argument(int is_short, char*** parg, char** popt)
{
  char* value = NULL;

  /* find required parameter */
  if(is_short) {
    (*popt)++;
    if( **popt !=0 ) {
      value = *popt;
    }
  }
  if(!value) {
    value = strchr(*popt, '=');
    if(value) {
      value++;
    } else {
      /* print error if required parameter not found */
      if( (*parg)[1] == 0 ) {
        return NULL;
      }
       (*parg)++;
       value = *(*parg);
    }
  }
  /* skip the rest of the current program argument parsing */
  *popt = value + strlen(value) - 1;
  return value;
}

/**
 * Process given command line option
 *
 * @param o option to process
 * @param is_short non zero if option is in its short form
 * @param parg pointer to the first unprocessed program argument
 * @param popt option name
 */
static void process_option(cmdline_opt_t *o, int is_short, char*** parg, char** popt)
{
  char* cur_op = *popt;
  /* NOTE: the function can be rewrote using iterator <param,opt,is_short> */
  char* value = NULL;
  
  if(is_param_required(o->type)) {
    value = retrive_option_argument(is_short, parg, popt);
    if(!value) {
      /* print error if required parameter not found */
      log_msg(PROGRAM_NAME ": argument is required for option ");
      if(is_short) log_msg("-%c\n", *cur_op); 
      else log_msg("--%s\n", cur_op);
      rsh_exit(2);
    }
  }
  process_option_2(&opt, o, value);
}

/**
 * Parse command line of the program.
 *
 * @param argv a command line
 * @return 1 on success
 */
static int parse_cmdline(char *argv[])
{
  char **arg;
  char *op;
  cmdline_opt_t* t;
  
  /* initialize options structure */
  opt.find_max_depth = -1;
  opt.argv = argv + 1;
  IF_WINDOWS(win_convert_cmdline_to_internal_encoding());

  for(arg = opt.argv; *arg != NULL; arg++) {
    op = *arg;
    /* check if argument is an option ("-" considered as alias for stdin) */
    if(*op != '-' || op[1] == 0) {
      break;
    }
    /* is it a long option? */
    if(op[1] == '-') {
      size_t len;
      if(op[2] == 0) {
        arg++;
        break;
      }
      op += 2;
      /* search for '=' */
      len = strcspn(op, "=");
      for(t = cmdline_opt; t->type; t++) {
        if(strncmp(op, t->long_name, len) == 0 && strlen(t->long_name) == len) {
          break;
        }
      }
      if(!t->type) {
        log_msg(PROGRAM_NAME ": unknown option: --%s\n", op);
        rsh_exit(2);
      }
      /* process a long option */
      process_option(t, 0, &arg, &op);
    } else {
      for(op++; *op; op++) {
        for(t = cmdline_opt; t->type && *op != t->short1 && *op != t->short2; t++);
        if(!t->type) {
          log_msg(PROGRAM_NAME ": unknown option: -%c\n", *op);
          rsh_exit(2);
        }
        /* process a short option */
        process_option(t, 1, &arg, &op);
      }
    }
  }
  
  /* the rest of arguments are treated as filenames */
  opt.files = arg;

#ifdef _WIN32
  /* strip '\','/' symbols from the end of filepath */
  for(arg = opt.files; *arg != NULL; arg++) {
    char *p = *arg;
    int i = (int)strlen(p)-1;
    for(; i > 0 && IS_PATH_SEPARATOR(p[i]) && p[i-1] != ':'; i--) p[i] = 0;
  }
#endif

  /* copy formating options from config if not specified at cmd line */
  if(!opt.printf && !opt.template_file && !opt.sum_flags && !opt.fmt) {
    opt.printf = conf_opt.printf;
    opt.template_file = conf_opt.template_file;
  }

  if(!opt.printf && !opt.template_file) {
    if(!opt.fmt) opt.fmt = conf_opt.fmt;
    if(!opt.sum_flags) opt.sum_flags = conf_opt.sum_flags;
  }

  if(!opt.mode)  opt.mode = conf_opt.mode;
  opt.flags |= conf_opt.flags; /* copy all non-sum options */

  if( opt.files_accept == 0 )  opt.files_accept = conf_opt.files_accept;
  if( opt.crc_accept == 0 )    opt.crc_accept = conf_opt.crc_accept;
  if( opt.embed_crc_delimiter == 0 ) opt.embed_crc_delimiter = conf_opt.embed_crc_delimiter;
  if( !opt.path_separator ) opt.path_separator = conf_opt.path_separator;
  if( opt.find_max_depth < 0 ) opt.find_max_depth = conf_opt.find_max_depth;
  if( opt.flags & OPT_EMBED_CRC ) opt.sum_flags |= RHASH_CRC32;
  if( opt.openssl_mask == 0 ) opt.openssl_mask = conf_opt.openssl_mask;

#ifdef _WIN32
  if( (opt.flags & OPT_ENCODING) == 0 ) opt.flags |= (conf_opt.flags & OPT_ENCODING);
  if( (opt.flags & OPT_ENCODING) == 0 ) opt.flags |= OPT_UTF8;
#endif

  /* set defaults */
  if(opt.embed_crc_delimiter == 0) opt.embed_crc_delimiter = " ";
  return 1;
}

/**
 * Search for config file.
 * 
 * @return the relative path to config file
 */
static const char* find_conf_file(void)
{
  struct rsh_stat_struct st;
  char *dir1, *path;

#ifndef _WIN32 /* Linux/Unix part */
  /* first check for $HOME/.rhashrc file */
  if( (dir1 = getenv("HOME")) ) {
    path = make_path(dir1, ".rhashrc");
    if(rsh_stat(path, &st) >= 0) {
      vector_add_ptr(opt.mem, path);
      return (conf_opt.config_file = path);
    }
    free(path);
  }
  /* then check for global config */
  if(rsh_stat( (path = "/etc/rhashrc"), &st) >= 0) {
    return (conf_opt.config_file = path);
  }

#else /* _WIN32 */
# define HOME_CONF_FILE "rhashrc"

  /* first check for the %APPDATA%\RHash\rhashrc config */
  if( (dir1 = getenv("APPDATA")) ) {
    dir1 = make_path(dir1, "RHash");
    vector_add_ptr(opt.mem, path = make_path(dir1, HOME_CONF_FILE));
    free(dir1);
    if(rsh_stat(path, &st) >= 0) {
      return (conf_opt.config_file = path);
    }
  }

  /* then check for %HOMEDRIVE%%HOMEPATH%\rhashrc */
  /* note that %USERPROFILE% is generally not a user home dir */
  if( (dir1 = getenv("HOMEDRIVE")) && (path = getenv("HOMEPATH"))) {
    dir1 = make_path(dir1, path);
    vector_add_ptr(opt.mem, path = make_path(dir1, HOME_CONF_FILE));
    free(dir1);
    if(rsh_stat(path, &st) >= 0) {
      return (conf_opt.config_file = path);
    }
  }
#endif /* _WIN32 */

  return (conf_opt.config_file = NULL); /* config not found */
}

/**
 * Parse config file of the program.
 *
 * @return 1 on success, 0 on fail
 */
static int read_config(void)
{
#define LINE_BUF_SIZE 2048
  char buf[LINE_BUF_SIZE];
  FILE* fd;

  /* initialize conf_opt and opt structures */
  memset(&conf_opt, 0, sizeof(opt));
  conf_opt.find_max_depth = -1;

  if(!find_conf_file()) return 1;

  fd = fopen(conf_opt.config_file, "r");
  if(!fd) return 1;

  while( fgets(buf, LINE_BUF_SIZE, fd) ) {
    size_t index;
    cmdline_opt_t* t;
    char* line = str_trim(buf);
    char  *name, *value;

    if(*line == 0 || IS_COMMENT(*line)) continue;

    /* search for '=' */
    index = strcspn(line, "=");
    if(line[index] == 0) {
      log_msg("%s: can't parse line: %s\n", conf_opt.config_file, line);
      continue;
    }
    line[index] = 0;
    name = str_trim(line);

    for(t = cmdline_opt; t->type; t++) {
      if(strcmp(name, t->long_name) == 0) {
        break;
      }
    }

    if(!t->type) {
      log_msg("%s: unknown option: %s\n", conf_opt.config_file, line);
      continue;
    }

    value = str_trim(line+index+1);

    /* process a long option */
    if(is_param_required(t->type)) {
      vector_add_ptr(opt.mem, (value = rsh_strdup(value)));;
    } else {
      static const char* strings[] = {"1", "on", "yes", 0};
      const char** cmp;
      for(cmp = strings; *cmp && strcmp(value, *cmp); cmp++);
      if(*cmp == 0) continue;
    }

    process_option_2(&conf_opt, t, value);
  }
  fclose(fd);

  return 1;
}

/**
 * Try to detect hash sums options from program name.
 *
 * @param progName the program name
 */
static void set_default_sums_flags(const char* progName)
{
  char *buf;
  int res = 0;

  /* remove directory name from path */
  const char* p = strrchr(progName, '/');
  if(p) progName = p+1;
#ifdef _WIN32
  p = strrchr(progName, '\\');
  if(p) progName = p+1;
#endif

  /* convert progName to lowercase */
  buf = str_tolower(progName);

  if(strstr(buf, "crc32")) res |= RHASH_CRC32;
  if(strstr(buf, "md4"))   res |= RHASH_MD4;
  if(strstr(buf, "md5"))   res |= RHASH_MD5;
  if(strstr(buf, "sha1"))  res |= RHASH_SHA1;
  if(strstr(buf, "sha256")) res |= RHASH_SHA256;
  if(strstr(buf, "sha512")) res |= RHASH_SHA512;
  if(strstr(buf, "sha224")) res |= RHASH_SHA224;
  if(strstr(buf, "sha384")) res |= RHASH_SHA384;
  if(strstr(buf, "tiger")) res |= RHASH_TIGER;
  if(strstr(buf, "tth"))   res |= RHASH_TTH;
  if(strstr(buf, "btih"))  res |= RHASH_BTIH;
  if(strstr(buf, "aich"))  res |= RHASH_AICH;
  if(strstr(buf, "gost"))  res |= RHASH_GOST;
  if(strstr(buf, "gost-cryptopro"))  res |= RHASH_GOST_CRYPTOPRO;
  if(strstr(buf, "has160"))  res |= RHASH_HAS160;
  if(strstr(buf, "ripemd160"))  res |= RHASH_RIPEMD160;
  if(strstr(buf, "whirlpool"))  res |= RHASH_WHIRLPOOL;
  if(strstr(buf, "edonr256"))   res |= RHASH_EDONR256;
  if(strstr(buf, "edonr512"))   res |= RHASH_EDONR512;
  if(strstr(buf, "snefru256"))  res |= RHASH_SNEFRU128;
  if(strstr(buf, "snefru128"))  res |= RHASH_SNEFRU256;
  if(strstr(buf, "ed2k-link") || strstr(buf, "ed2k-hash")) res |= OPT_ED2K_LINK;
  else if(strstr(buf, "ed2k")) res |= RHASH_ED2K;

  if(strstr(buf, "sfv") && opt.fmt == 0) opt.fmt = FMT_SFV;
  if(strstr(buf, "magnet") && opt.fmt == 0) opt.fmt = FMT_MAGNET;
  
  free(buf);

  /* change program flags only if opt.sum_flags was not set */
  if(!opt.sum_flags) {
    opt.sum_flags = (res ? res : (opt.fmt == FMT_MAGNET ? RHASH_TTH | RHASH_ED2K | RHASH_AICH : RHASH_CRC32));
  }
}

/**
 * Log a message and exit the program.
 *
 * @param msg the message to log
 */
static void die(const char* msg)
{
  log_msg(msg);
  rsh_exit(2);
}

/**
 * Destroy a parsed options object.
 *
 * @param o pointer to the options object to destroy.
 */
void options_destroy(struct options_t* o)
{
  file_mask_free(o->files_accept);
  file_mask_free(o->crc_accept);
  vector_free(o->cmd_vec);
  vector_free(o->mem);
}

/**
 * Parse command line options.
 *
 * @param argv program arguments
 */
void read_options(char *argv[])
{
  unsigned ff; /* formating flags */
  memset(&opt, 0, sizeof(opt));
  opt.mem = vector_new_simple();

  read_config();
  parse_cmdline(argv);

  IF_WINDOWS(setup_console());
  setup_output(); /* setup program output */

  if((opt.flags & OPT_VERBOSE) && conf_opt.config_file) {
    /* note that the first log_msg call must be made after setup_output() */
    log_msg("Config file: %s\n", (conf_opt.config_file ? conf_opt.config_file : "None"));
  }

  /* check that no more than one program mode specified */
  if(opt.mode & (opt.mode - 1)) {
    die(PROGRAM_NAME ": incompatible program modes\n");
  }

  ff = (opt.printf ? 1 : 0) | (opt.template_file ? 2 : 0) | (opt.fmt ? 4 : 0);
  if((opt.fmt & (opt.fmt - 1)) || (ff & (ff - 1))) {
    die(PROGRAM_NAME ": too many formating options\n");
  }

  if(opt.mode & MODE_TORRENT) opt.sum_flags |= RHASH_BTIH;

  if(!opt.crc_accept) opt.crc_accept = file_mask_new_from_list(".sfv");

  if(opt.openssl_mask) rhash_transmit(RMSG_SET_OPENSSL_MASK, 0, opt.openssl_mask, 0);

  set_default_sums_flags(argv[0]); /* detect default hashes from program name */
}
